# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.
import re
from io import BytesIO, StringIO
from lxml import etree
from gs123.conversion import BarcodeConverter


def convert_xml_string(data: str,
                       company_prefix_length: int = 6,
                       serial_number_length: int = 12):
    """
    Converts all matching barcode patterns in an xml string.
    :param data: The data with barcode data
    :param company_prefix_length: The company prefix length
    :param serial_number_length: The serial number length
    :return: The data string with the converted values inserted.
    """
    if isinstance(data, bytes):
        elements = etree.iterparse(BytesIO(data),
                                   events=('start', 'end',),
                                   remove_comments=True)
    elif isinstance(data, str):
        elements = etree.iterparse(BytesIO(data.encode('utf-8')),
                                   events=('start', 'end',),
                                   remove_comments=True)

    with StringIO('') as output_file:
        _parse_xml(company_prefix_length, elements,
                   serial_number_length)
        return (etree.tostring(elements.root))


def convert_xml_file(file_path: str,
                     output_file_path: str,
                     company_prefix_length: int = 6,
                     serial_number_length: int = 12):
    """
    Converts an inbound XML file into an outbound XML file with all of the
    barcodes converted to EPC URN values.
    :param file_path: The file to parse.
    :param output_file_path: The new file to create.
    :param company_prefix_length: The length of the company prefix in the
    barcodes. Default is 6.
    :param serial_number_length: The serial number length.  Default is 12.
    :return: None.
    """
    elements = etree.iterparse(file_path, events=('start', 'end',),
                               remove_comments=True)
    with open(output_file_path, 'wb+') as output_file:
        _parse_xml(company_prefix_length, elements,
                   serial_number_length)
        output_file.write(etree.tostring(elements.root))
        output_file.flush()


def _parse_xml(company_prefix_length, elements,
               serial_number_length, converter_type=BarcodeConverter):
    for event, element in elements:
        try:
            if event == 'end':
                bc = converter_type(
                    element.text,
                    company_prefix_length=company_prefix_length,
                    max_serial_number_length=serial_number_length
                )
                element.text = bc.epc_urn
            for name, value in element.items():
                bc = converter_type(
                    value,
                    company_prefix_length=company_prefix_length,
                    max_serial_number_length=serial_number_length
                )
                element.set(name, bc.epc_urn)

        except BarcodeConverter.BarcodeNotValid:
            pass



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.

from gs123.xml_conversion import BarcodeConverter, convert_xml_string
from gs123.conversion import URNConverter
from quartet_capture import models
from quartet_capture.rules import Step, RuleContext


class BaseConversionClass(Step):
    """
    Base class for the conversion steps in this module.
    """

    def __init__(self, db_task: models.Task, **kwargs):
        super().__init__(db_task, **kwargs)
        self._declared_parameters = {
            "Context Key": "The context key where the data "
                           "to be converted can "
                           "be located.  Default is NUMBER_RESPONSE",
            "Use Context Key": "Whether or not to look in the context using "
                               "the context key for XML to convert. Default "
                               "is True.  If false, the step will "
                               "look in the inbound data for barcodes "
                               "to convert.",
            "Serial Number Length": "The length of the serial number field if "
                                    "the barcode data does not have delimited "
                                    "app identifiers with "
                                    "parenthesis and also "
                                    "has lot and expiry fields. "
                                    "Default is 12.",
            "Company Prefix Length": "The length of the company prefix.  "
                                     "Default is 6."
        }
        self.company_prefix_length, \
        self.context_key, \
        self.serial_number_length, \
        self.use_context_key = self._get_parameter_values()

    @property
    def declared_parameters(self):
        return self._declared_parameters

    def on_failure(self):
        pass

    def _get_parameter_values(self):
        """
        Checks all the step parameters for configured or default values
        and rerturns.
        :return: A tuple containing company prefix length, the rule context
        key to examine, the serial number length, and a boolean value
        which determines whether the context key should be used to look
        up data to covert or if the step should convert the data parameter
        passed into the execute function.
        """
        company_prefix_length = self.get_or_create_parameter(
            'Company Prefix Length',
            '6',
            self.declared_parameters['Company Prefix Length']
        )
        serial_number_length = self.get_or_create_parameter(
            'Serial Number Length',
            '12',
            self.declared_parameters['Serial Number Length']
        )
        context_key = self.get_or_create_parameter(
            'Context Key',
            'NUMBER_RESPONSE',
            self.declared_parameters['Context Key']
        )
        use_context_key = self.get_or_create_parameter(
            'Use Context Key',
            'False',
            self.declared_parameters['Use Context Key']
        )
        use_context_key = ('true' == use_context_key.lower())
        if use_context_key:
            self.info('Using context key %s' % context_key)
        else:
            self.info('Using the rule data.')
        return (
            company_prefix_length, context_key,
            serial_number_length, use_context_key
        )


class ListBarcodeConversionStep(BaseConversionClass):
    """
    If a list is passed into the rule as data or via the configurable
    step parameter "Context Key" in which the data is provided in the
    rule context, will convert any items
    in the list to valid URNs.
    """

    def __init__(self, db_task: models.Task, **kwargs):
        super().__init__(db_task, **kwargs)
        self._declared_parameters["Property"] = \
            "The name of the property to access on the instance.  Check the" \
            " properties on the BarcodeConverter class for available options."
        self.prop_name = self.get_parameter('Property', 'epc_urn')

    def execute(self, data, rule_context: RuleContext):
        self.info('Task parameters: %s',
                  str(self.get_task_parameters(rule_context)))
        to_process = data or rule_context.context.get(self.context_key)
        if isinstance(to_process, list):
            converted = [self.convert(item) for item in to_process]
            if data:
                self.info('Inbound data was converted.  Returning back '
                          'to rule.')
                return converted
            else:
                self.info('The information in context key %s was converted.',
                          self.context_key)
                rule_context[self.context_key] = converted
        else:
            self.warning('No list data was provided for conversion.')

    def convert(self, data):
        """
        Will convert the data parameter to a urn value and return.
        Override this to return a different value from the BarcodeConverter.
        :param data: The barcode value to convert.
        :return: An EPC URN based on the inbound data.
        """
        prop_val = BarcodeConverter(
            data,
            self.company_prefix_length,
            self.serial_number_length
        ).__getattribute__(self.prop_name)
        return prop_val if isinstance(prop_val, str) else prop_val()


class ListURNConversionStep(ListBarcodeConversionStep):
    """
    Converts URN values into barcode values.  Using the Property step parameter
    you can specify what property on the `ListBarcodeConversionStep` object
    to return.
    """

    def __init__(self, db_task: models.Task, **kwargs):
        super().__init__(db_task, **kwargs)
        self._declared_parameters["Property"] = \
            "The name of the property to access on the instance.  Check the" \
            " properties on the URNConverter class for available options."
        self.prop_name = self.get_parameter('Property', 'get_barcode_value')

    def convert(self, data):
        """
        Will convert the data parameter to a urn value and return.
        Override this to return a different value from the BarcodeConverter.
        :param data: The barcode value to convert.
        :return: An EPC URN based on the inbound data.
        """
        prop_val = URNConverter(
            data
        ).__getattribute__(self.prop_name)
        return prop_val if isinstance(prop_val, str) else prop_val()


class XMLBarcodeConversionStep(BaseConversionClass):
    """
    Will look in the rule context for XML with barcode data in it that the
    gs123 can convert, convert it and replace it with the new converted (to
    urn) data.
    """

    def execute(self, data, rule_context: RuleContext):
        if self.use_context_key:
            barcode_xml = rule_context.context.get(self.context_key, None)
            if isinstance(barcode_xml, bytes):
                barcode_xml = barcode_xml.decode('utf-8')
        else:
            barcode_xml = data
        if barcode_xml:
            converted_data = convert_xml_string(
                barcode_xml,
                int(self.company_prefix_length),
                int(self.serial_number_length)
            ).decode('utf-8')
            self.info('Barcode data has been filtered and replaced where '
                      'possible.')
            if not self.use_context_key and converted_data:
                return converted_data
            else:
                rule_context.context[self.context_key] = converted_data
        else:
            self.warning('No XML was found in the Rule Context under the '
                         'context key %s or the rule had no inbound data.'
                         % self.context_key)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.
import math

def calculate_check_digit(data):
    """
    Given a numeric string of n length, will calculate a GS1 compliant
    check digit according to instructions supplied here:
    https://www.gs1.org/how-calculate-check-digit-manually
    :param data: A numeric string of data.
    :return: The input string with the appropriate check digit applied.
    """
    sum = 0
    factors = [1,3] if len(data) & 1 else[3, 1]
    for i in range(len(data)):
        sum += int(data[i]) * factors[0] if i & 1 else int(data[i]) * factors[1]
    return "%s%s" % (data, str(int(math.ceil(sum/10.0)) * 10 - sum))



# -*- coding: utf-8
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.

from __future__ import unicode_literals, absolute_import
import os
import django

DEBUG = True
USE_TZ = True

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "u_6k^p*oqxbq*h+sbm(kc=iqw&vk6_@gp#!d%p=zqh7a&(^=_o"
DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    }
}


REST_FRAMEWORK = {
    'DEFAULT_PERMISSION_CLASSES': (
        'rest_framework.permissions.IsAuthenticated',
        'rest_framework.permissions.DjangoModelPermissions'
    ),
}

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sites",
    "quartet_epcis",
    "quartet_capture.apps.QuartetCaptureConfig",
]

SITE_ID = 1

if django.VERSION >= (1, 10):
    MIDDLEWARE = ()
else:
    MIDDLEWARE_CLASSES = ()

MEDIA_ROOT = '/tmp/'
#
# import logging
# logging.basicConfig(
#     level=logging.DEBUG,
#     handlers=[logging.StreamHandler()]
# )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.

import os
import sys
import click

try:
    from gs123.xml_conversion import convert_xml_file
except ImportError:
    sys.path.append(os.path.join('../',os.path.dirname(__file__)))
    from gs123.xml_conversion import convert_xml_file


@click.command()
@click.option(
    '-i', '--input-file',
    help='An input file with barcode data to convert'
)
@click.option(
    '-o', '--output-file',
    help='The output file for converted data'
)
def main(input_file, output_file):
    """Console script for gs123."""
    input_file = os.path.abspath(input_file)
    output_file = os.path.abspath(output_file)
    convert_xml_file(input_file, output_file)
    return 0


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.

# To edit this go here: https://regex101.com/r/v9CJm9/3

import re

_NUMERIC_GS1_01_21_OPTIONAL_17_10 = (r'^\(01\)(?P<gtin14>[0-9]{14})'
                                     r'\(21\)(?P<serial_number>[0-9, A-Z]{1,20})'
                                     r'(\(17\)(?P<expiration_date>[0-9]{6})?'
                                     r'\(10\)(?P<lot>[a-zA-Z0-9]{1,20}))?$')

# to edit https://regex101.com/r/TSraS8/1
_NO_PARENS_NUMERIC_GS1_01_21 = (r'^01(?P<gtin14>[0-9]{14})21(?P<serial_number>'
                                r'[0-9,A-Z]{1,20})$')

# to edit https://regex101.com/r/H0fHK8/2/
# the {%serial_number_length%} must be replaced at run-time by the
# calling application
_NO_PARENS_NUMERIC_GS1_01_21_OPTIONAL_17_10 = \
    (r'^01(?P<gtin14>[0-9]{14})'
     r'21(?P<serial_number>[0-9,A-Z]{{%serial_number_length%}})'
     r'(17(?P<expiration_date>(\d{6}))'
     r'10(?P<lot>[\x21-\x22\x25-\x2F\x30-\x39\x41-\x5A\x5F\x61-\x7A]{0,20}))?$')

NUMERIC_GS1_01_21_OPTIONAL_17_10 = re.compile(
    _NUMERIC_GS1_01_21_OPTIONAL_17_10)

NO_PARENS_NUMERIC_GS1_01_21 = re.compile(_NO_PARENS_NUMERIC_GS1_01_21)

_ALPHA_01_21_GTIN_NO_PARENS = r'^01(?P<gtin14>[0-9]{14})21(?P<serial_number>[0-9,a-z,A-Z]{10,20})$'
ALPHA_01_21_GTIN_NO_PARENS = re.compile(_ALPHA_01_21_GTIN_NO_PARENS)


# https://regex101.com/r/RkpSY6/3/
# universal, works with just about every pattern
_SGTIN_SN_10_13_ALPHA = r'^(01|\(01\))(?P<gtin14>[0-9]{14})(21|\(21\))(?P<serial_number>[0-9,a-z,A-Z]{10,13})((17|\(17\))(?P<expiration_date>\d{6}))?((10|\(10\))(?P<lot>[\x21-\x22\x25-\x2F\x30-\x39\x3A-\x3F\x41-\x5A\x5F\x61-\x7A]{0,20}))?$'
SGTIN_SN_10_13_ALPHA = re.compile(_SGTIN_SN_10_13_ALPHA)


def get_no_parens_numeric_gs1_01_21_optional_17_10(serial_number_length=12):
    """
    Will return a matching compiled regular expression to use in barcode values
    that have a fixed length serial number field but not parenthesis in the
    app identifiers.
    :param serial_number_length: The length of the fixed serial number field.
    :return: A compiled regular expression.
    """
    pattern = _NO_PARENS_NUMERIC_GS1_01_21_OPTIONAL_17_10.replace(
        '{%serial_number_length%}', str(serial_number_length)
    )
    return re.compile(pattern)


# https://regex101.com/r/H0fHK8/5
_FNC1_SERIAL = r'^01(?P<gtin14>[0-9]{14})21(?P<serial_number>[0-9,A-Z]*?(\x1d\b))(17(?P<expiration_date>[0-9]{6})10(?P<lot>[\x21-\x22\x25-\x2F\x30-\x39\x41-\x5A\x5F\x61-\x7A]{0,20}))?'
FNC1_SERIAL = re.compile(_FNC1_SERIAL)

# https://regex101.com/r/ivtuux/1/
_SSCC = r'^(00|\(00\))?(?P<sscc18>\d{18})$'
SSCC = re.compile(_SSCC)

#https://regex101.com/r/PsHODE/1
_GTIN14 = r'^(01|\(01\))?(?P<gtin14>\d{14})$'
GTIN14 = re.compile(_GTIN14)

# https://regex101.com/r/wjN6lC/1/
NO_PARENS_NUMERIC_GS1_01_21_IN_DOC = r'01(?P<gtin14>[0-9]{14})21(?P<serial_number>[0-9,A-Z]{1,20})'

# https://regex101.com/r/mVhm8D/2
SGTIN_URN = r'urn:epc:id:sgtin:(?P<company_prefix>[0-9]{1,10}).(?P<item_reference>[0-9]{1,10}).(?P<serial_number>[0-9a-zA-Z]{1,20})'

# https://regex101.com/r/ik2XH2/1/
SSCC_URN = r'urn:epc:id:sscc:(?P<company_prefix>[0-9]{4,10}).(?P<serial_number>[0-9]{1,10})'

urn_patterns = [
    re.compile(SGTIN_URN),
    re.compile(SSCC_URN)
]

patterns = [
    GTIN14,
    SSCC
]

def match_pattern(barcode_val: str, max_serial_number_length=14):
    """
    Will use the regular expressions in this module to find common barcode
    components and return the match.  For an example of how to use this see
    the `gs123.conversion.BarcodeConverter._populate` function.
    :param barcode_val: The value to match.
    :return: A regex match or none.
    """
    match = False
    matches = []
    barcode_val = str(barcode_val)
    if barcode_val.startswith('(01)'):
        match = SGTIN_SN_10_13_ALPHA.match(
            barcode_val
        )
    elif barcode_val.startswith('01'):
        if len(barcode_val) <= 18 + max_serial_number_length:
            match = ALPHA_01_21_GTIN_NO_PARENS.match(
                barcode_val
            )
    elif barcode_val.startswith('(00)') or barcode_val.startswith('00'):
        match = SSCC.match(
            barcode_val
        )
    if not match:
        if not match:
            match = NO_PARENS_NUMERIC_GS1_01_21.match(
                barcode_val
            )
        if not match:
            match = get_no_parens_numeric_gs1_01_21_optional_17_10(
            ).match(
                barcode_val
            )
        if not match:
            match = FNC1_SERIAL.match(
                barcode_val
            )
    return match



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""
import os, sys, re

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

requirements = ['Click>=6.0', ]

setup_requirements = []

test_requirements = []

def get_version(*file_paths):
    """Retrieves the version from quartet_integrations/__init__.py"""
    filename = os.path.join(os.path.dirname(__file__), *file_paths)
    version_file = open(filename).read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string.')


version = get_version("gs123", "__init__.py")


if sys.argv[-1] == 'publish':
    try:
        import wheel
        print("Wheel version: ", wheel.__version__)
    except ImportError:
        print('Wheel library missing. Please run "pip install wheel"')
        sys.exit()
    os.system('python setup.py sdist upload')
    os.system('python setup.py bdist_wheel upload')
    sys.exit()

if sys.argv[-1] == 'tag':
    print("Tagging the version on git:")
    os.system("git tag -a %s -m 'version %s'" % (version, version))
    os.system("git push --tags")
    sys.exit()


setup(
    author="Rob Magee",
    author_email='slab@serial-lab.com',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    description="Set of python-based tools to convert various GS1 formats "
                "to and from others.",
    entry_points={
        'console_scripts': [
            'gs123=gs123.cli:main',
        ],
    },
    install_requires=requirements,
    license="GNU General Public License v3",
    long_description=readme,
    include_package_data=True,
    keywords='gs123',
    name='gs123',
    packages=find_packages(include=['gs123']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://gitlab.com/serial-lab/gs123',
    version=version,
    zip_safe=False,
)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.

import re
from gs123 import regex
from gs123.check_digit import calculate_check_digit

FNC1 = '\x1D'


class BarcodeConverter:
    """
    Converts barcode values to URN values.
    """

    def __init__(self, barcode_val: str,
                 company_prefix_length: int,
                 max_serial_number_length: int = 14):
        """
        Initializes a new conversion class from a serialized GTIN in either
        01...21...17...10 or (01)...(21)...(17...(10) or 01...21 or
        (01)...(21) formats.  Once the inbound barcode value is parsed,
        the constituent components of both the GTIN-14 and the overall barcode
        itself are available via the properties of the class.  See the unit
        tests in this package for usage examples or consult the included
        documentation's *jupyter notebook*.

        :param barcode_val: The barcode value to convert.
        :param company_prefix_length: The company prefix.
        :param max_serial_number_length: The length of the serial number if
        the app identifiers do not have parenthesis and there are 17 and 10
        fields after the serial number field.
        """
        self._company_prefix_length = company_prefix_length
        self._max_serial_number_length = max_serial_number_length
        self._gtin14 = None
        self._sscc18 = None
        self._extension_digit = None
        self._serial_number = None
        self._padded_serial_number = None
        self._lot = None
        self._expiration_date = None
        self._indicator_digit = None
        self._company_prefix = None
        self._item_reference = None
        self._check_digit = None
        self._sgtin_pattern = 'urn:epc:id:sgtin:{0}.{1}{2}.{3}'
        self._sscc_pattern = 'urn:epc:id:sscc:{0}.{1}{2}'
        self._is_gtin = False
        if not barcode_val:
            raise self.BarcodeNotValid('No barcode was present.')
        match = regex.match_pattern(barcode_val, max_serial_number_length)
        if match:
            self._populate(match)
        else:
            raise self.BarcodeNotValid(
                'The barcode %s was not valid against the regular expressions '
                'available in the module.' % barcode_val
            )

    @property
    def sscc18(self):
        return self._sscc18

    @property
    def extension_digit(self):
        return self._extension_digit

    @property
    def gtin14(self) -> str:
        """
        Returns the GTIN 14 for the class.
        :return: A string value with GTIN 14
        """
        return self._gtin14

    @property
    def padded_serial_number(self) -> str:
        """
        Returns as Serial Number, as-is.
        :return: Serial Number as string.
        """
        return self._padded_serial_number

    @property
    def serial_number(self) -> str:
        """
        Returns the serial number as a string with any leading zeros stripped.
        To get a non-stripped or non-altered serial number, use the
        `serial_number_field` property.
        :return: Serial number as string.
        """
        return self._serial_number.lstrip("0")

    @property
    def serial_number_field(self) -> str:
        """
        For the `serial_number` property, any leading zeros will be stripped
        from the serial number field if they are leading
        zeros.  This function will return the leading zeros if any were
        present.
        :return:
        """
        return self._serial_number

    @property
    def lot(self) -> str:
        """
        Returns the lot number as a string or None if no lot number was in
        the barcode.
        :return: String or None
        """
        return self._lot

    @property
    def expiration_date(self) -> str:
        """
        Returns the expiration date section of the barcode as a string or
        none if no expiration date was included.
        :return: Expiry as string or None
        """
        return self._expiration_date

    @property
    def indicator_digit(self) -> str:
        return self._gtin14[0]

    @property
    def item_reference(self) -> str:
        """
        The item reference number as taken from the barcode value.
        """
        return self._gtin14[self._company_prefix_length + 1:13]

    @property
    def company_prefix(self) -> str:
        """
        The company prefix as taken from the barcode value.
        :return: String
        """
        if self._gtin14:
            ret = self._gtin14[1:self._company_prefix_length + 1]
        else:
            ret = self._sscc18[1:self._company_prefix_length + 1]
        return ret

    @property
    def check_digit(self) -> str:
        """
        The original check digit supplied in the barcode.
        :return:
        """
        if self._gtin14:
            ret = self._gtin14[13]
        else:
            ret = self._sscc18[17]
        return ret

    @property
    def epc_urn(self) -> str:
        """
        Returns a GS1 EPC URN value for the given barcode.
        :return: String GS1 EPC URN for the barcode.
        """
        if self._gtin14:
            ret = self._sgtin_pattern.format(
                self.company_prefix,
                self.indicator_digit,
                self.item_reference,
                self.serial_number
            )
        else:
            ret = self._sscc_pattern.format(
                self.company_prefix,
                self.extension_digit,
                self.serial_number_field
            )
        return ret

    @property
    def padded_epc_urn(self) -> str:
        """
          If the Barcode is a GTIN-14, this will return the Serial Number 'as-is'. Meaning the
          Serial Number will not have leading Zeros removed.
          :return: String GS1 EPC URN for the Barcode
        """

        if self._gtin14:
            ret = self._sgtin_pattern.format(
                self.company_prefix,
                self.indicator_digit,
                self.item_reference,
                self.padded_serial_number
            )
        else:
            return self.epc_urn
        return ret

    @property
    def epc_urn_fixed_serial(self) -> str:
        """
        Will return a (technically not-valid) EPC URN value with the
        serial number as it was reported in the barcode- meaning if there
        was any left zero padding, that padding will be included in the
        serial number field.
        :return:
        """
        return self._sgtin_pattern.format(
            self.company_prefix,
            self.indicator_digit,
            self.item_reference,
            self.padded_serial_number
        )

    def _populate(self, match) -> None:
        """
        If there is an inbound barcode that produces a match,
        this function will populate the class with all of the
        necessary converted properties.
        :param match: The Match object.
        :return: None
        """
        group_dict = match.groupdict()
        self._gtin14 = group_dict.get('gtin14')
        if self.gtin14:
            self._serial_number = str(
                group_dict['serial_number'].strip('\x1d'))
            self._padded_serial_number = self._serial_number
            self._expiration_date = group_dict.get('expiration_date')
            self._lot = group_dict.get('lot')
        else:  # this means we have an SSCC
            self._sscc18 = group_dict['sscc18']
            self._serial_number = self._sscc18[
                                  1 + int(self._company_prefix_length):17]
            self._extension_digit = self._sscc18[0]

    class BarcodeNotValid(BaseException):
        pass


class URNConverter(BarcodeConverter):
    """
    Converts an EPC urn to a valid barcode.
    """

    def __init__(self, urn_value: str):
        """
        Will convert the urn_value into a valid barcode.
        :param urn_value: A urn to convert.
        :param lot: If the lot is supplied an app identifier field of 10 will
        be added to the barcode
        :param expiration: If an expiration date value is supplied, this will
        be added to the barcode as a 17 field/app identifier value.
        """
        self._gtin14 = None
        self.is_sgtin = True
        for pattern in regex.urn_patterns:
            match = pattern.match(
                urn_value
            )
            if match:
                self._populate(match, urn_value)

    def _populate(self, match, urn_value: str):
        """
        Will populate the class parameters based on the match where
        applicable.
        :param match: The regular expression match
        :return: None
        """
        groups = match.groupdict()
        self._serial_number = str(groups.get('serial_number'))
        self._padded_serial_number = self._serial_number
        self._company_prefix = groups.get('company_prefix')
        if self._company_prefix:
            self._company_prefix_length = len(self._company_prefix)
        if urn_value.startswith('urn:epc:id:sgtin:'):
            self._handle_sgtin_urn(groups, urn_value)
        elif urn_value.startswith('urn:epc:id:sscc:'):
            self._handle_sscc_urn(groups, urn_value)

    def _handle_sscc_urn(self, groups, urn_value):
        """
        Extracts sscc info from the urn value and the regex groups.
        :param groups: The groups that were matched.
        :param urn_value: The original urn.
        :return: None
        """
        self.is_sgtin = False
        serial_number = groups.get('serial_number')
        self._extension_digit = serial_number[:1]
        self._serial_number = serial_number[1:]
        barcode = '%s%s' % (self._extension_digit, self._company_prefix)
        padding_length = 17 - len(barcode)
        serial_number = self._serial_number.zfill(padding_length)
        barcode = '%s%s' % (barcode, serial_number)
        self._sscc18 = calculate_check_digit(barcode)

    def _handle_sgtin_urn(self, groups, urn_value):
        """
        Extracts sgtin info from the urn value and the regex groups.
        :param groups: The groups that were matched.
        :param urn_value: The original urn.
        :return: None
        """
        item_reference = groups.get('item_reference')
        self._item_reference = item_reference[1:]
        self._indicator_digit = item_reference[:1]
        self._gtin14 = '{0}{1}{2}'.format(
            self._indicator_digit,
            self._company_prefix,
            self._item_reference
        )
        if len(self._gtin14) != 13:
            raise URNNotValid(
                'The urn %s is not valid. The parts representing the '
                'GTIN-14 should be 13 digits in length total.'
                % urn_value
            )
        self._gtin14 = calculate_check_digit(self._gtin14)

    def get_barcode_value(self, lot=None, expiration=None,
                          insert_control_char=False, parenthesis=False,
                          serial_number_padding=False,
                          serial_number_length=12,
                          padding_character='0'):
        """
        Returns the properly formatted barcode value.
        :param lot: pass this in if you'd like to add a lot (10)
        app identifier and lot value
        :param expiration: Pass this in if you'd like to add an expiration (17)
        and an expiration value
        :param insert_control_char: Set to true if you'd like an FNC1
        delimiter inserted after the serial number field.
        :param parenthesis: Set to true if you'd like parenthesis around all
        app identifiers.
        :param serial_number_padding: Set to true to pad serial numbers in
        SGTIN barcodes.
        :param serial_number_length: Set the length of the serial number length
        if you plan to pad it.
        :param padding_character: The character to pad the serial number with-
        default is zero
        :return: The properly formatted barcode string.
        """
        if self.is_sgtin:
            return self._get_gtin_barcode_val(
                lot=lot, expiration=expiration,
                insert_control_char=insert_control_char,
                parenthesis=parenthesis,
                serial_number_padding=serial_number_padding,
                serial_number_length=serial_number_length,
                padding_character=padding_character
            )
        else:
            return self._get_sscc_barcode_val(parenthesis=parenthesis)

    def _get_gtin_barcode_val(self, lot=None, expiration=None,
                              insert_control_char=False, parenthesis=False,
                              serial_number_padding=False,
                              serial_number_length=12,
                              padding_character='0') -> str:
        """
        Returns the barcode value for this instance.
        :return: A string representing the barcode.
        """
        if serial_number_padding:
            serial_number = self.serial_number.rjust(serial_number_length,
                                                     padding_character)
        else:
            serial_number = self.serial_number
        format_string = '(01){0}(21){1}' if parenthesis else '01{0}21{1}'
        barcode = format_string.format(
            self.gtin14,
            serial_number
        )
        if insert_control_char:
            barcode = '%s%s' % (barcode, FNC1)
        if expiration:
            if len(expiration) > 6:
                raise InvalidFieldDataError(
                    'The length of the expiration date must be six characters '
                    'long in YYMMDD format per GS1 standards.'
                )
            format_string = '%s(17)%s' if parenthesis else '%s17%s'
            barcode = format_string % (barcode, expiration)
        if lot:
            format_string = '%s(10)%s' if parenthesis else '%s10%s'
            barcode = format_string % (barcode, lot)
        return barcode

    def _get_sscc_barcode_val(self, parenthesis=False) -> str:
        format_string = '(00)%s' if parenthesis else '00%s'
        barcode = '%s%s' % (self.extension_digit, self.company_prefix)
        padding_length = 17 - len(barcode)
        serial_number = self._serial_number.zfill(padding_length)
        barcode = '%s%s' % (barcode, serial_number)
        return format_string % calculate_check_digit(barcode)


class URNNotValid(Exception):
    """
    Raised by instances when the inbound urn value is malformed.
    """
    pass


class InvalidFieldDataError(Exception):
    """
    Raised if inbound data is bad.
    """
    pass



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.

import os

import django
from django.test import TestCase
from django.db.utils import IntegrityError

os.environ['DJANGO_SETTINGS_MODULE'] = 'tests.test_settings'
django.setup()

from quartet_capture import models
from quartet_capture.rules import Rule

from gs123.conversion import BarcodeConverter, URNConverter, FNC1
from gs123.xml_conversion import convert_xml_file, convert_xml_string
from gs123.check_digit import calculate_check_digit


class TestGs123(TestCase):
    """Tests for `gs123` package."""

    def test_01_21_no_parens_leading_zeros(self):
        converter = BarcodeConverter(
            '011234567890123421003456789012',
            6
        )
        self.assertEqual(converter.gtin14, '12345678901234')
        self.assertEqual(converter.serial_number_field, '003456789012')
        self.assertEqual(converter.company_prefix, '234567')
        self.assertEqual(converter.indicator_digit, '1')
        self.assertEqual(converter.check_digit, '4')
        self.assertEqual(converter.item_reference, '890123')
        self.assertEqual(converter.epc_urn,
                         'urn:epc:id:sgtin:234567.1890123.3456789012')

    def test_01_21_no_parens(self):
        converter = BarcodeConverter(
            '01003005406390512110008344372010',
            6
        )
        self.assertEqual(converter.gtin14, '00300540639051')
        self.assertEqual(converter.serial_number, '10008344372010')
        self.assertEqual(converter.company_prefix, '030054')
        self.assertEqual(converter.indicator_digit, '0')
        self.assertEqual(converter.check_digit, '1')
        self.assertEqual(converter.item_reference, '063905')
        self.assertEqual(converter.epc_urn,
                         'urn:epc:id:sgtin:030054.0063905.10008344372010')

    def test_01_21_no_parens_12_digit(self):
        converter = BarcodeConverter(
            '013030054063905221100000001307',
            6
        )
        self.assertEqual(converter.gtin14, '30300540639052')
        self.assertEqual(converter.serial_number, '100000001307')
        self.assertEqual(converter.company_prefix, '030054')
        self.assertEqual(converter.indicator_digit, '3')
        self.assertEqual(converter.check_digit, '2')
        self.assertEqual(converter.item_reference, '063905')
        self.assertEqual(converter.epc_urn,
                         'urn:epc:id:sgtin:030054.3063905.100000001307')

    def test_01_21_with_parens(self):
        converter = BarcodeConverter(
            '(01)12345678901234(21)123456789012',
            6
        )
        self.assertEqual(converter.gtin14, '12345678901234')
        self.assertEqual(converter.serial_number, '123456789012')
        self.assertEqual(converter.company_prefix, '234567')
        self.assertEqual(converter.indicator_digit, '1')
        self.assertEqual(converter.check_digit, '4')
        self.assertEqual(converter.item_reference, '890123')
        self.assertEqual(converter.epc_urn,
                         'urn:epc:id:sgtin:234567.1890123.123456789012')

    def test_sscc_18(self):
        converter = BarcodeConverter(
            '00050991510019004305',
            company_prefix_length=7
        )
        self.assertEqual(converter.company_prefix, "5099151")
        self.assertEqual(converter.serial_number_field, "001900430")

    def test_01_21_17_10_with_parens(self):
        converter = BarcodeConverter(
            '(01)00312345678901(21)000000000001(17)191231(10)ABC123',
            6
        )
        self.assertEqual(converter.gtin14, '00312345678901')
        self.assertEqual(converter.serial_number_field, '000000000001')
        self.assertEqual(converter.serial_number, '1')
        self.assertEqual(converter.company_prefix, '031234')
        self.assertEqual(converter.indicator_digit, '0')
        self.assertEqual(converter.check_digit, '1')
        self.assertEqual(converter.item_reference, '567890')
        self.assertEqual(converter.epc_urn,
                         'urn:epc:id:sgtin:031234.0567890.1')

    def test_01_21_17_10_no_parens(self):
        converter = BarcodeConverter(
            '0100312345678901210000000000011719123110ABC123',
            company_prefix_length=6,
            max_serial_number_length=12
        )
        self.assertEqual(converter.gtin14, '00312345678901')
        self.assertEqual(converter.serial_number, '1')
        self.assertEqual(converter.serial_number_field, '000000000001')
        self.assertEqual(converter.company_prefix, '031234')
        self.assertEqual(converter.indicator_digit, '0')
        self.assertEqual(converter.check_digit, '1')
        self.assertEqual(converter.item_reference, '567890')
        self.assertEqual(converter.epc_urn,
                         'urn:epc:id:sgtin:031234.0567890.1')
        self.assertEqual(converter.lot, 'ABC123')
        self.assertEqual(converter.expiration_date, '191231')

    def test_01_21_17_10_no_parens_FNC1(self):
        converter = BarcodeConverter(
            '010031234567890121000000000000001\x1D1719123110ABC123\x1D',
            company_prefix_length=6,
        )
        self.assertEqual(converter.gtin14, '00312345678901')
        self.assertEqual(converter.serial_number_field, '000000000000001')
        self.assertEqual(converter.serial_number, '1')
        self.assertEqual(converter.company_prefix, '031234')
        self.assertEqual(converter.indicator_digit, '0')
        self.assertEqual(converter.check_digit, '1')
        self.assertEqual(converter.item_reference, '567890')
        self.assertEqual(converter.epc_urn,
                         'urn:epc:id:sgtin:031234.0567890.1')
        self.assertEqual(converter.lot, 'ABC123')
        self.assertEqual(converter.expiration_date, '191231')

    def test_01_21_17_10_no_parens_FNC1_padded(self):
        converter = BarcodeConverter(
            '010031234567890121000000000000001\x1D1719123110ABC123\x1D',
            company_prefix_length=6,
        )
        self.assertEqual(converter.gtin14, '00312345678901')
        self.assertEqual(converter.serial_number_field, '000000000000001')
        self.assertEqual(converter.serial_number, '1')
        self.assertEqual(converter.padded_serial_number, '000000000000001')
        self.assertEqual(converter.company_prefix, '031234')
        self.assertEqual(converter.indicator_digit, '0')
        self.assertEqual(converter.check_digit, '1')
        self.assertEqual(converter.item_reference, '567890')
        self.assertEqual(converter.padded_epc_urn,
                         'urn:epc:id:sgtin:031234.0567890.000000000000001')

        self.assertEqual(converter.epc_urn,
                         'urn:epc:id:sgtin:031234.0567890.1')

        self.assertEqual(converter.lot, 'ABC123')
        self.assertEqual(converter.expiration_date, '191231')

    def test_sscc18_with_parens(self):
        converter = BarcodeConverter(
            '(00)012345612345678907',
            6
        )
        self.assertEqual(converter.extension_digit, '0')
        self.assertEqual(converter.sscc18, '012345612345678907')
        self.assertEqual(converter.serial_number, '1234567890')
        self.assertEqual(converter.check_digit, '7')
        self.assertEqual(converter.company_prefix, '123456')
        self.assertEqual(
            converter.epc_urn,
            'urn:epc:id:sscc:123456.01234567890'
        )

    def test_sscc18_without_parens_7_cp_len(self):
        converter = BarcodeConverter(
            '00012345612345678907',
            7, 12
        )
        self.assertEqual(converter.extension_digit, '0')
        self.assertEqual(converter.sscc18, '012345612345678907')
        self.assertEqual(converter.serial_number, '234567890')
        self.assertEqual(converter.check_digit, '7')
        self.assertEqual(converter.company_prefix, '1234561')
        self.assertEqual(
            converter.epc_urn,
            'urn:epc:id:sscc:1234561.0234567890'
        )

    def test_sscc18_without_parens(self):
        converter = BarcodeConverter(
            '00012345612345678907',
            6
        )
        self.assertEqual(converter.extension_digit, '0')
        self.assertEqual(converter.sscc18, '012345612345678907')
        self.assertEqual(converter.serial_number, '1234567890')
        self.assertEqual(converter.check_digit, '7')
        self.assertEqual(converter.company_prefix, '123456')
        self.assertEqual(
            converter.epc_urn,
            'urn:epc:id:sscc:123456.01234567890'
        )

    def test_bad_barcode(self):
        with self.assertRaises(BarcodeConverter.BarcodeNotValid):
            converter = BarcodeConverter(
                '012392348439', 6
            )

    def test_check_digits(self):
        self.assertEqual(
            calculate_check_digit("0099999999998"),
            "00999999999986"
        )
        self.assertEqual(
            calculate_check_digit("12345678901234567"),
            "123456789012345675"
        )
        self.assertEqual(
            calculate_check_digit("123456789012"),
            "1234567890128"
        )

    def test_file_conversion(self):
        curpath = os.path.join(os.path.dirname(__file__),
                               'data/serialnumbers.xml')
        output_file_path = os.path.join(os.path.dirname(__file__),
                                        'data/converted.xml')
        convert_xml_file(curpath, output_file_path, company_prefix_length=6,
                         serial_number_length=10)

    def test_sscc_file_conversion(self):
        curpath = os.path.join(os.path.dirname(__file__),
                               'data/ssccs.xml')
        output_file_path = os.path.join(os.path.dirname(__file__),
                                        'data/sscc_converted.xml')
        convert_xml_file(curpath, output_file_path, company_prefix_length=6,
                         serial_number_length=10)

    def test_string_conversion(self):
        data = """<?xml version='1.0' encoding='UTF-8'?>
<S:Envelope xmlns:S="http://schemas.xmlsoap.org/soap/envelope/">
    <S:Body>
        <ns2:serialNumbersRequestResponse xmlns:ns2="urn:test:soap">
            <SNResponse>
                <ReceivingSystem>0344444000006</ReceivingSystem>
                <SendingSystem>0344444000006</SendingSystem>
                <ActionCode>C</ActionCode>
                <EncodingType>SGTIN</EncodingType>
                <IDType>GS1_SER</IDType>
                <ObjectKey>
                    <Name>GTIN</Name>
                    <Value>00377713112102</Value>
                </ObjectKey>
                <RandomizedNumberList>
                    <SerialNo>0100377713112102211RFXVHNPA111</SerialNo>
                    <SerialNo>01003777131121022114R2FANWAG12</SerialNo>
                    <SerialNo>0100377713112102212FWA6AVK7614</SerialNo>
                    <SerialNo>0100377713112102212NN3NG5VK415</SerialNo>
                    <SerialNo>01003777131121022119KNN3H4A145</SerialNo>
                    <SerialNo>01003777131121022125P4N3X8NP45</SerialNo>
                    <SerialNo>01003777131121022116326N1GFV75</SerialNo>
                    <SerialNo>010037771311210221148NNK9N7488</SerialNo>
                    <SerialNo>01003777131121022115WANPT8KR34</SerialNo>
                    <SerialNo>01003777131121022113CK6FRH7R88</SerialNo>
                    <SerialNo>0100377713112102211X769VGH1G7J</SerialNo>
                    <SerialNo>010037771311210221325NV1T32FSD</SerialNo>
                    <SerialNo>01003777131121022117F4VTPWR5CV</SerialNo>
                    <SerialNo>0100377713112102212P5W5R9WRGED</SerialNo>
                    <SerialNo>0100377713112102211NK693FK75FF</SerialNo>
                    <SerialNo>0100377713112102212F397C3455LM</SerialNo>
                    <SerialNo>0100377713112102212F76HPVFF5ED</SerialNo>
                    <SerialNo>0100377713112102212CWRCFTTPTEW</SerialNo>
                    <SerialNo>0100377713112102211N3F9PTP14DF</SerialNo>
                    <SerialNo>010037771311210221245RV96KFHJK</SerialNo>
                </RandomizedNumberList>
            </SNResponse>
        </ns2:serialNumbersRequestResponse>
    </S:Body>
</S:Envelope>"""
        ret = convert_xml_string(data, company_prefix_length=6)
        print(ret)

    def test_bytes_step(self):
        """
        Test the XMLBarcodeConversion step.
        :return: None
        """
        data = b"""<?xml version='1.0' encoding='UTF-8'?>
        <S:Envelope xmlns:S="http://schemas.xmlsoap.org/soap/envelope/">
            <S:Body>
                <ns2:serialNumbersRequestResponse xmlns:ns2="urn:test:soap">
                    <SNResponse>
                        <ReceivingSystem>0344444000006</ReceivingSystem>
                        <SendingSystem>0344444000006</SendingSystem>
                        <ActionCode>C</ActionCode>
                        <EncodingType>SGTIN</EncodingType>
                        <IDType>GS1_SER</IDType>
                        <ObjectKey>
                            <Name>GTIN</Name>
                            <Value>00377713112102</Value>
                        </ObjectKey>
                        <RandomizedNumberList>
                            <SerialNo>0100377713112102211RFXVHNPA111</SerialNo>
                            <SerialNo>01003777131121022114R2FANWAG12</SerialNo>
                            <SerialNo>0100377713112102212FWA6AVK7614</SerialNo>
                            <SerialNo>0100377713112102212NN3NG5VK415</SerialNo>
                            <SerialNo>01003777131121022119KNN3H4A145</SerialNo>
                            <SerialNo>01003777131121022125P4N3X8NP45</SerialNo>
                            <SerialNo>01003777131121022116326N1GFV75</SerialNo>
                            <SerialNo>010037771311210221148NNK9N7488</SerialNo>
                            <SerialNo>01003777131121022115WANPT8KR34</SerialNo>
                            <SerialNo>01003777131121022113CK6FRH7R88</SerialNo>
                            <SerialNo>0100377713112102211X769VGH1G7J</SerialNo>
                            <SerialNo>010037771311210221325NV1T32FSD</SerialNo>
                            <SerialNo>01003777131121022117F4VTPWR5CV</SerialNo>
                            <SerialNo>0100377713112102212P5W5R9WRGED</SerialNo>
                            <SerialNo>0100377713112102211NK693FK75FF</SerialNo>
                            <SerialNo>0100377713112102212F397C3455LM</SerialNo>
                            <SerialNo>0100377713112102212F76HPVFF5ED</SerialNo>
                            <SerialNo>0100377713112102212CWRCFTTPTEW</SerialNo>
                            <SerialNo>0100377713112102211N3F9PTP14DF</SerialNo>
                            <SerialNo>010037771311210221245RV96KFHJK</SerialNo>
                        </RandomizedNumberList>
                    </SNResponse>
                </ns2:serialNumbersRequestResponse>
            </S:Body>
        </S:Envelope>"""
        db_rule, db_task, db_step = self._create_rule()
        c_rule = Rule(db_task.rule, db_task)
        c_rule.context.context['NUMBER_RESPONSE'] = data
        c_rule.execute('')
        self.assertTrue('urn:epc:id:sgtin:037771.0311210.1RFXVHNPA111' in
                        c_rule.context.context['NUMBER_RESPONSE'])
        c_rule.context.context['NUMBER_RESPONSE'] = data
        c_rule.execute('')
        self.assertTrue('urn:epc:id:sgtin:037771.0311210.1RFXVHNPA111' in
                        c_rule.context.context['NUMBER_RESPONSE'])

    def test_string_step(self):
        """
        Test the XMLBarcodeConversion step.
        :return: None
        """
        data = """<?xml version='1.0' encoding='UTF-8'?>
        <S:Envelope xmlns:S="http://schemas.xmlsoap.org/soap/envelope/">
            <S:Body>
                <ns2:serialNumbersRequestResponse xmlns:ns2="urn:test:soap">
                    <SNResponse>
                        <ReceivingSystem>0344444000006</ReceivingSystem>
                        <SendingSystem>0344444000006</SendingSystem>
                        <ActionCode>C</ActionCode>
                        <EncodingType>SGTIN</EncodingType>
                        <IDType>GS1_SER</IDType>
                        <ObjectKey>
                            <Name>GTIN</Name>
                            <Value>00377713112102</Value>
                        </ObjectKey>
                        <RandomizedNumberList>
                            <SerialNo>0100377713112102211RFXVHNPA111</SerialNo>
                            <SerialNo>01003777131121022114R2FANWAG12</SerialNo>
                            <SerialNo>0100377713112102212FWA6AVK7614</SerialNo>
                            <SerialNo>0100377713112102212NN3NG5VK415</SerialNo>
                            <SerialNo>01003777131121022119KNN3H4A145</SerialNo>
                            <SerialNo>01003777131121022125P4N3X8NP45</SerialNo>
                            <SerialNo>01003777131121022116326N1GFV75</SerialNo>
                            <SerialNo>010037771311210221148NNK9N7488</SerialNo>
                            <SerialNo>01003777131121022115WANPT8KR34</SerialNo>
                            <SerialNo>01003777131121022113CK6FRH7R88</SerialNo>
                            <SerialNo>0100377713112102211X769VGH1G7J</SerialNo>
                            <SerialNo>010037771311210221325NV1T32FSD</SerialNo>
                            <SerialNo>01003777131121022117F4VTPWR5CV</SerialNo>
                            <SerialNo>0100377713112102212P5W5R9WRGED</SerialNo>
                            <SerialNo>0100377713112102211NK693FK75FF</SerialNo>
                            <SerialNo>0100377713112102212F397C3455LM</SerialNo>
                            <SerialNo>0100377713112102212F76HPVFF5ED</SerialNo>
                            <SerialNo>0100377713112102212CWRCFTTPTEW</SerialNo>
                            <SerialNo>0100377713112102211N3F9PTP14DF</SerialNo>
                            <SerialNo>010037771311210221245RV96KFHJK</SerialNo>
                        </RandomizedNumberList>
                    </SNResponse>
                </ns2:serialNumbersRequestResponse>
            </S:Body>
        </S:Envelope>"""
        db_rule, db_task, db_step = self._create_rule()
        c_rule = Rule(db_task.rule, db_task)
        c_rule.context.context['NUMBER_RESPONSE'] = data.encode('utf-8')
        c_rule.execute('')
        self.assertTrue('urn:epc:id:sgtin:037771.0311210.1RFXVHNPA111' in
                        c_rule.context.context['NUMBER_RESPONSE'])

    def test_params(self):
        data = """<?xml version='1.0' encoding='UTF-8'?>
        <S:Envelope xmlns:S="http://schemas.xmlsoap.org/soap/envelope/">
            <S:Body>
                <ns2:serialNumbersRequestResponse xmlns:ns2="urn:test:soap">
                    <SNResponse>
                        <ReceivingSystem>0344444000006</ReceivingSystem>
                        <SendingSystem>0344444000006</SendingSystem>
                        <ActionCode>C</ActionCode>
                        <EncodingType>SGTIN</EncodingType>
                        <IDType>GS1_SER</IDType>
                        <ObjectKey>
                            <Name>GTIN</Name>
                            <Value>00377713112102</Value>
                        </ObjectKey>
                        <RandomizedNumberList>
                            <SerialNo>0100377713112102211RFXVHNPA1111712311910LOT777</SerialNo>
                            <SerialNo>01003777131121022114R2FANWAG121712311910LOT777</SerialNo>
                            <SerialNo>0100377713112102212FWA6AVK76141712311910LOT777</SerialNo>
                            <SerialNo>0100377713112102212NN3NG5VK4151712311910LOT777</SerialNo>
                            <SerialNo>01003777131121022119KNN3H4A1451712311910LOT777</SerialNo>
                            <SerialNo>01003777131121022125P4N3X8NP451712311910LOT777</SerialNo>
                            <SerialNo>01003777131121022116326N1GFV751712311910LOT777</SerialNo>
                            <SerialNo>010037771311210221148NNK9N74881712311910LOT777</SerialNo>
                            <SerialNo>01003777131121022115WANPT8KR341712311910LOT777</SerialNo>
                            <SerialNo>01003777131121022113CK6FRH7R881712311910LOT777</SerialNo>
                            <SerialNo>0100377713112102211X769VGH1G7J1712311910LOT777</SerialNo>
                            <SerialNo>010037771311210221325NV1T32FSD1712311910LOT777</SerialNo>
                            <SerialNo>01003777131121022117F4VTPWR5CV1712311910LOT777</SerialNo>
                            <SerialNo>0100377713112102212P5W5R9WRGED1712311910LOT777</SerialNo>
                            <SerialNo>0100377713112102211NK693FK75FF1712311910LOT777</SerialNo>
                            <SerialNo>0100377713112102212F397C3455LM1712311910LOT777</SerialNo>
                            <SerialNo>0100377713112102212F76HPVFF5ED1712311910LOT777</SerialNo>
                            <SerialNo>0100377713112102212CWRCFTTPTEW1712311910LOT777</SerialNo>
                            <SerialNo>0100377713112102211N3F9PTP14DF1712311910LOT777</SerialNo>
                            <SerialNo>010037771311210221245RV96KFHJK1712311910LOT777</SerialNo>
                        </RandomizedNumberList>
                    </SNResponse>
                </ns2:serialNumbersRequestResponse>
            </S:Body>
        </S:Envelope>"""
        db_rule, db_task, db_step = self._create_rule()
        company_prefix_length = models.StepParameter.objects.create(
            name='Company Prefix Length',
            value='7',
            step=db_step
        )
        serial_number_length = models.StepParameter.objects.create(
            name='Serial Number Length',
            value='10',
            step=db_step
        )
        c_rule = Rule(db_task.rule, db_task)
        c_rule.context.context['NUMBER_RESPONSE'] = data.encode('utf-8')
        c_rule.execute('')
        self.assertTrue('urn:epc:id:sgtin:0377713.011210.1RFXVHNPA111' in
                        c_rule.context.context['NUMBER_RESPONSE'])
        c_rule = Rule(db_task.rule, db_task)
        c_rule.context.context['NUMBER_RESPONSE'] = ''
        c_rule.execute('')
        self.assertTrue('urn:epc:id:sgtin:0377713.011210.1RFXVHNPA111' not in
                        c_rule.context.context['NUMBER_RESPONSE'])

    def test_parse_rule_data(self):
        data = """<?xml version='1.0' encoding='UTF-8'?>
        <S:Envelope xmlns:S="http://schemas.xmlsoap.org/soap/envelope/">
            <S:Body>
                <ns2:serialNumbersRequestResponse xmlns:ns2="urn:test:soap">
                    <SNResponse>
                        <ReceivingSystem>0344444000006</ReceivingSystem>
                        <SendingSystem>0344444000006</SendingSystem>
                        <ActionCode>C</ActionCode>
                        <EncodingType>SGTIN</EncodingType>
                        <IDType>GS1_SER</IDType>
                        <ObjectKey>
                            <Name>GTIN</Name>
                            <Value>00377713112102</Value>
                        </ObjectKey>
                        <RandomizedNumberList>
                            <SerialNo>0100377713112102211RFXVHNPA1111712311910LOT777</SerialNo>
                            <SerialNo>01003777131121022114R2FANWAG121712311910LOT777</SerialNo>
                            <SerialNo>0100377713112102212FWA6AVK76141712311910LOT777</SerialNo>
                            <SerialNo>0100377713112102212NN3NG5VK4151712311910LOT777</SerialNo>
                            <SerialNo>01003777131121022119KNN3H4A1451712311910LOT777</SerialNo>
                            <SerialNo>01003777131121022125P4N3X8NP451712311910LOT777</SerialNo>
                            <SerialNo>01003777131121022116326N1GFV751712311910LOT777</SerialNo>
                            <SerialNo>010037771311210221148NNK9N74881712311910LOT777</SerialNo>
                            <SerialNo>01003777131121022115WANPT8KR341712311910LOT777</SerialNo>
                            <SerialNo>01003777131121022113CK6FRH7R881712311910LOT777</SerialNo>
                            <SerialNo>0100377713112102211X769VGH1G7J1712311910LOT777</SerialNo>
                            <SerialNo>010037771311210221325NV1T32FSD1712311910LOT777</SerialNo>
                            <SerialNo>01003777131121022117F4VTPWR5CV1712311910LOT777</SerialNo>
                            <SerialNo>0100377713112102212P5W5R9WRGED1712311910LOT777</SerialNo>
                            <SerialNo>0100377713112102211NK693FK75FF1712311910LOT777</SerialNo>
                            <SerialNo>0100377713112102212F397C3455LM1712311910LOT777</SerialNo>
                            <SerialNo>0100377713112102212F76HPVFF5ED1712311910LOT777</SerialNo>
                            <SerialNo>0100377713112102212CWRCFTTPTEW1712311910LOT777</SerialNo>
                            <SerialNo>0100377713112102211N3F9PTP14DF1712311910LOT777</SerialNo>
                            <SerialNo>010037771311210221245RV96KFHJK1712311910LOT777</SerialNo>
                        </RandomizedNumberList>
                    </SNResponse>
                </ns2:serialNumbersRequestResponse>
            </S:Body>
        </S:Envelope>"""
        db_rule, db_task, db_step = self._create_rule()
        models.StepParameter.objects.create(
            name='Company Prefix Length',
            value='7',
            step=db_step
        )
        models.StepParameter.objects.create(
            name='Serial Number Length',
            value='10',
            step=db_step
        )
        sp = models.StepParameter.objects.get(
            name='Use Context Key',
        )
        sp.step = db_step
        sp.value = 'False'
        sp.save()
        c_rule = Rule(db_task.rule, db_task)
        data = c_rule.execute(data)

    def test_parse_urn_data(self):
        rule, task, conversion_step = self._create_urn_rule()
        c_rule = Rule(task.rule, task)
        vals = [
            'urn:epc:id:sgtin:077722.0011210.12CW68RW6G',
            'urn:epc:id:sgtin:077722.0011210.11X8KN3H4W',
            'urn:epc:id:sgtin:077722.0011210.29A1X6F8KR',
            'urn:epc:id:sgtin:077722.0011210.15684ART9V',
            'urn:epc:id:sgtin:077722.0011210.2KK8NT4264'
        ]
        c_rule.execute(vals)
        print(c_rule.data)

    def _create_rule(self):
        db_rule = models.Rule()
        db_rule.name = 'xml_barcode_conversion'
        db_rule.description = 'XML Barcode conversion step rule.'
        db_rule.save()
        rp = models.RuleParameter(name='test name', value='test value',
                                  rule=db_rule)
        rp.save()
        # create a new step
        conversion_step = models.Step()
        conversion_step.name = 'parse-barcodes'
        conversion_step.description = 'Convert barcodes to urns.'
        conversion_step.order = 1
        conversion_step.step_class = 'gs123.steps.XMLBarcodeConversionStep'
        conversion_step.rule = db_rule
        conversion_step.save()

        ucc_param = models.StepParameter.objects.create(
            name='Use Context Key',
            value='True',
            step=conversion_step,
        )

        db_task = models.Task(
            rule=db_rule,
            status='QUEUED',
        )
        return db_rule, db_task, conversion_step

    def _create_urn_rule(self):
        db_rule = models.Rule()
        db_rule.name = 'List URN Conversion'
        db_rule.description = 'Converts a list of URN values to barcodes.'
        db_rule.save()

        conversion_step = models.Step()
        conversion_step.name = 'List URN Parser'
        conversion_step.description = 'Parse a list of URN values.'
        conversion_step.order = 1
        conversion_step.step_class = 'gs123.steps.ListURNConversionStep'
        conversion_step.rule = db_rule
        conversion_step.save()

        db_task = models.Task(
            rule=db_rule,
            status='QUEUED',
        )
        return db_rule, db_task, conversion_step

    def test_urn_to_barcode(self):
        urnc = URNConverter(
            'urn:epc:id:sgtin:0377713.011210.1RFXVHNPA111'
        )
        self.assertEqual(
            urnc.get_barcode_value(),
            '0100377713112109211RFXVHNPA111'
        )

    def test_parens(self):
        urnc = URNConverter(
            'urn:epc:id:sgtin:0377713.011210.1RFXVHNPA111'
        )
        self.assertEqual(
            urnc.get_barcode_value(parenthesis=True),
            '(01)00377713112109(21)1RFXVHNPA111'
        )

    def test_lot_batch(self):
        urnc = URNConverter(
            'urn:epc:id:sgtin:0377713.011210.1RFXVHNPA111'
        )
        self.assertEqual(
            urnc.get_barcode_value(
                lot='ABC123',
                expiration='181231'
            ),
            '0100377713112109211RFXVHNPA1111718123110ABC123'
        )

    def test_lot_batch_parens(self):
        urnc = URNConverter(
            'urn:epc:id:sgtin:0377713.011210.1RFXVHNPA111'
        )
        self.assertEqual(
            urnc.get_barcode_value(
                lot='ABC123',
                expiration='181231',
                parenthesis=True
            ),
            '(01)00377713112109(21)1RFXVHNPA111(17)181231(10)ABC123'
        )

    def test_padding(self):
        urnc = URNConverter(
            'urn:epc:id:sgtin:0377713.011210.1RFXVHNPA111'
        )
        self.assertEqual(
            urnc.get_barcode_value(
                serial_number_length=14,
                serial_number_padding=True
            ),
            '010037771311210921001RFXVHNPA111',
        )

    def test_fnc1(self):
        urnc = URNConverter(
            'urn:epc:id:sgtin:0377713.011210.1RFXVHNPA111'
        )
        self.assertEqual(
            urnc.get_barcode_value(
                lot='ABC123',
                expiration='181231',
                insert_control_char=True
            ),
            '0100377713112109211RFXVHNPA111{0}1718123110ABC123'.format(FNC1)
        )

    def test_serial_number(self):
        urnc = URNConverter(
            'urn:epc:id:sgtin:0377713.011210.1RFXVHNPA111'
        )
        self.assertEqual(
            urnc.serial_number,
            '1RFXVHNPA111'
        )

    def _fixture_teardown(self):
        try:
            super()._fixture_teardown()
        except IntegrityError:
            pass

